@@ -1,44 +1,26 @@
-# Copyright (c) 2022, Riverbank Computing Limited
-# All rights reserved.
-#
-# This copy of SIP is licensed for use under the terms of the SIP License
-# Agreement.  See the file LICENSE for more details.
-#
-# This copy of SIP may also used under the terms of the GNU General Public
-# License v2 or v3 as published by the Free Software Foundation which can be
-# found in the files LICENSE-GPL2 and LICENSE-GPL3 included in this package.
-#
-# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
-# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
-# POSSIBILITY OF SUCH DAMAGE.
+# SPDX-License-Identifier: BSD-2-Clause
+
+# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>
 
 
 from dataclasses import dataclass, field
 from enum import auto, Enum
-from typing import Any, List, Optional, Union
+from typing import Any, Dict, List, Optional, Union
+
+from .scoped_name import ScopedName
 
 
 class AccessSpecifier(Enum):
     """ The class access specifiers. """
 
-    # TODO: Change the values to auto() once the C code has been replaced.
-
     # Private access.
-    PRIVATE = 0x04
+    PRIVATE = auto()
 
     # Protected access.
-    PROTECTED = 0x02
+    PROTECTED = auto()
 
     # Public access.
-    PUBLIC = 0x01
+    PUBLIC = auto()
 
 
 class ArgumentType(Enum):
@@ -225,45 +207,38 @@ class ArgumentType(Enum):
 class ArrayArgument(Enum):
     """ The array support provided by an argument. """
 
-    # TODO: Change the values to auto() once the C code has been replaced.
-
     # /Array/ was specified.
-    ARRAY = 0
+    ARRAY = auto()
 
     # /ArraySize/ was specified.
-    ARRAY_SIZE = 1
+    ARRAY_SIZE = auto()
 
     # The argument provides no array support.
-    NONE = 2
+    NONE = auto()
 
 
 class ClassKey(Enum):
     """ The key that identifies a particular type of class. """
 
-    # TODO: Change the values to auto() once the C code has been replaced.
-
     # A class.
-    CLASS = 0
+    CLASS = auto()
 
     # A struct.
-    STRUCT = 1
+    STRUCT = auto()
 
     # A union.
-    UNION = 2
+    UNION = auto()
 
 
 class DocstringFormat(Enum):
     """ The formatting applied to the text of the docstring. """
 
-    # TODO: Change the values to auto() once the C code has been replaced.
-
-    # The signature is appended to the docstring.
     # Any leading spaces common to all non-blank lines in the docstring are
     # removed.
-    DEINDENTED = 1
+    DEINDENTED = auto()
 
     # The docstring is used as it is specified in the .sip file.
-    RAW = 0
+    RAW = auto()
 
 
 class DocstringSignature(Enum):
@@ -272,504 +247,318 @@ class DocstringSignature(Enum):
     docstring then it applies to all the class's ctors.
     """
 
-    # TODO: Change the values to auto() once the C code has been replaced.
-
     # The signature is appended to the docstring.
-    APPENDED = 2
+    APPENDED = auto()
 
     # The signature is discard.
-    DISCARDED = 0
+    DISCARDED = auto()
 
     # The signature is prepended to the docstring.
-    PREPENDED = 1
+    PREPENDED = auto()
 
 
 class EnumBaseType(Enum):
     """ The different base types fo an enum. """
 
-    # TODO: Change the values to auto() once the C code has been replaced.
-
     # enum.Enum
-    ENUM = 0
+    ENUM = auto()
 
     # enum.Flag
-    FLAG = 1
+    FLAG = auto()
 
     # enum.IntEnum
-    INT_ENUM = 2
+    INT_ENUM = auto()
 
     # enum.IntFlag
-    INT_FLAG = 3
+    INT_FLAG = auto()
 
     # enum.IntEnum with unsigned values.
-    UNSIGNED_INT_ENUM = 4
+    UINT_ENUM = auto()
 
 
 class GILAction(Enum):
     """ The action to take with the GIL when calling C/C++ code. """
 
-    # TODO: Change the values to auto() once the C code has been replaced.
-
     # The default action.
-    DEFAULT = 0
+    DEFAULT = auto()
 
     # Hold the GIL.
-    HOLD = 1
+    HOLD = auto()
 
     # Release the GIL.
-    RELEASE = 2
+    RELEASE = auto()
 
 
 class IfaceFileType(Enum):
     """ The type of an interface file. """
 
-    # TODO: Change the values to auto() once the C code has been replaced.
+    # A class.
+    CLASS = auto()
 
     # An %Exception.
-    EXCEPTION = 0
+    EXCEPTION = auto()
 
     # A %MappedType.
-    MAPPED_TYPE = 1
+    MAPPED_TYPE = auto()
 
     # A namespace.
-    NAMESPACE = 2
-
-    # A class.
-    CLASS = 3
+    NAMESPACE = auto()
 
 
 class KwArgs(Enum):
     """ The level of support for passing argument as keyword arguments. """
 
-    # TODO: Change the values to auto() once the C code has been replaced.
-
     # All named arguments can be passed as keyword arguments.
-    ALL = 1
+    ALL = auto()
 
     # Keyword arguments are not supported.
-    NONE = 0
+    NONE = auto()
 
     # All named optional arguments (ie. those with a default value) can be
     # passed as keyword arguments.
-    OPTIONAL = 2
+    OPTIONAL = auto()
 
 
 class PyQtMethodSpecifier(Enum):
     """ The PyQt-specific method specifier. """
 
-    # TODO: Change the values to auto() once the C code has been replaced.
-
     # A signal.
-    SIGNAL = 0x10
+    SIGNAL = auto()
 
     # A slot.
-    SLOT = 0x08
+    SLOT = auto()
 
 
 class PySlot(Enum):
     """ The Python slots corresponding to entries in a type object. """
 
-    # TODO: Change the values to auto() once the C code has been replaced.
-
     # tp_str
-    STR = 0
+    STR = auto()
 
     # tp_as_number.nb_int
-    INT = 1
+    INT = auto()
 
     # tp_as_number.nb_float
-    FLOAT = 2
+    FLOAT = auto()
 
     # tp.as_mapping.mp_length and tp.as_sequence.sq_length
-    LEN = 3
+    LEN = auto()
 
     # tp.as_sequence.sq_contains
-    CONTAINS = 4
+    CONTAINS = auto()
 
     # tp_as_number.nb_add
-    ADD = 5
+    ADD = auto()
 
     # tp.as_sequence.sq_concat
-    CONCAT = 6
+    CONCAT = auto()
 
     # tp_as_number.nb_subtract
-    SUB = 7
+    SUB = auto()
 
     # tp_as_number.nb_multiply
-    MUL = 8
+    MUL = auto()
 
     # tp.as_sequence.sq_repeat
-    REPEAT = 9
+    REPEAT = auto()
 
     # tp_as_number.nb_remainder
-    MOD = 11
+    MOD = auto()
 
     # tp_as_number.nb_floor_divide
-    FLOORDIV = 12
+    FLOORDIV = auto()
 
     # tp_as_number.nb_true_divide
-    TRUEDIV = 13
+    TRUEDIV = auto()
 
     # tp_as_number.nb_and
-    AND = 14
+    AND = auto()
 
     # tp_as_number.nb_or
-    OR = 15
+    OR = auto()
 
     # tp_as_number.nb_xor
-    XOR = 16
+    XOR = auto()
 
     # tp_as_number.nb_lshift
-    LSHIFT = 17
+    LSHIFT = auto()
 
     # tp_as_number.nb_rshift
-    RSHIFT = 18
+    RSHIFT = auto()
 
     # tp_as_number.nb_inplace_add
-    IADD = 19
+    IADD = auto()
 
     # tp.as_sequence.sq_inplace_concat
-    ICONCAT = 20
+    ICONCAT = auto()
 
     # tp_as_number.nb_inplace_subtract
-    ISUB = 21
+    ISUB = auto()
 
     # tp_as_number.nb_inplace_multiply
-    IMUL = 22
+    IMUL = auto()
 
     # tp.as_sequence.sq_inplace_repeat
-    IREPEAT = 23
+    IREPEAT = auto()
 
     # tp_as_number.nb_inplace_remainder
-    IMOD = 25
+    IMOD = auto()
 
     # tp_as_number.nb_inplace_floor_divide
-    IFLOORDIV = 26
+    IFLOORDIV = auto()
 
     # tp_as_number.nb_inplace_true_divide
-    ITRUEDIV = 27
+    ITRUEDIV = auto()
 
     # tp_as_number.nb_inplace_and
-    IAND = 28
+    IAND = auto()
 
     # tp_as_number.nb_inplace_or
-    IOR = 29
+    IOR = auto()
 
     # tp_as_number.nb_inplace_xor
-    IXOR = 30
+    IXOR = auto()
 
     # tp_as_number.nb_inplace_lshift
-    ILSHIFT = 31
+    ILSHIFT = auto()
 
     # tp_as_number.nb_inplace_rshift
-    IRSHIFT = 32
+    IRSHIFT = auto()
 
     # tp_as_number.nb_invert
-    INVERT = 33
+    INVERT = auto()
 
     # tp_call
-    CALL = 34
+    CALL = auto()
 
     # tp.as_mapping.mp_subscript and tp.as_sequence.sq_item
-    GETITEM = 35
+    GETITEM = auto()
 
     # tp.as_mapping.mp_ass_subscript and tp.as_sequence.sq_ass_item
-    SETITEM = 36
+    SETITEM = auto()
 
     # tp.as_mapping.mp_ass_subscript and tp.as_sequence.sq_ass_item
-    DELITEM = 37
+    DELITEM = auto()
 
     # tp_richcompare (Py_LT)
-    LT = 38
+    LT = auto()
 
     # tp_richcompare (Py_LE)
-    LE = 39
+    LE = auto()
 
     # tp_richcompare (Py_EQ)
-    EQ = 40
+    EQ = auto()
 
     # tp_richcompare (Py_NE)
-    NE = 41
+    NE = auto()
 
     # tp_richcompare (Py_GT)
-    GT = 42
+    GT = auto()
 
     # tp_richcompare (Py_GE)
-    GE = 43
+    GE = auto()
 
     # tp_as_number.nb_bool
-    BOOL = 45
+    BOOL = auto()
 
     # tp_as_number.nb_negative
-    NEG = 46
+    NEG = auto()
 
     # tp_as_number.nb_positive
-    POS = 47
+    POS = auto()
 
     # tp_as_number.nb_absolute
-    ABS = 48
+    ABS = auto()
 
     # tp_repr
-    REPR = 49
+    REPR = auto()
 
     # tp_hash
-    HASH = 50
+    HASH = auto()
 
     # tp_as_number.nb_index
-    INDEX = 51
+    INDEX = auto()
 
     # tp_iter
-    ITER = 52
+    ITER = auto()
 
     # tp_iter_next
-    NEXT = 53
+    NEXT = auto()
 
     # tp_setattr
-    SETATTR = 54
+    SETATTR = auto()
 
     # Internal to the parser (implemented as tp_setattr)
-    DELATTR = 55
+    DELATTR = auto()
 
     # tp_as_number.nb_matrix_multiply
-    MATMUL = 56
+    MATMUL = auto()
 
     # tp_as_number.nb_inplace_matrix_multiply
-    IMATMUL = 57
+    IMATMUL = auto()
 
     # tp_as_async.am_await
-    AWAIT = 58
+    AWAIT = auto()
 
     # tp_as_async.am_aiter
-    AITER = 59
+    AITER = auto()
 
     # tp_as_async.am_anext
-    ANEXT = 60
+    ANEXT = auto()
 
 
 class QualifierType(Enum):
     """ The type of a qualifier used in %If/%End directives. """
 
-    # TODO: Change the values to auto() once the C code has been replaced.
-
     # The qualifier is a feature.
-    FEATURE = 2
+    FEATURE = auto()
 
     # The qualifier is a platform.
-    PLATFORM = 1
+    PLATFORM = auto()
 
     # The qualifier is part of a timeline.
-    TIME = 0
+    TIME = auto()
 
 
 class Transfer(Enum):
     """ The different types of ownership transfer. """
 
-    # TODO: Change the values to auto() once the C code has been replaced.
-
     # No transfer of ownership.
-    NONE = 0
+    NONE = auto()
 
     # /Transfer/ was specified.
-    TRANSFER = 1
+    TRANSFER = auto()
 
     # /TransferBack/ was specified.
-    TRANSFER_BACK = 2
+    TRANSFER_BACK = auto()
 
     # /TransferThis/ was specified.
-    TRANSFER_THIS = 3
+    TRANSFER_THIS = auto()
 
 
 class ValueType(Enum):
     """ The different types of a value in an expression. """
 
-    # TODO: Change the values to auto() once the C code has been replaced.
-
     # A quoted character.
-    QCHAR = 0
+    QCHAR = auto()
 
     # A string.
-    STRING = 1
+    STRING = auto()
 
     # A number.
-    NUMERIC = 2
+    NUMERIC = auto()
 
     # A floating point number.
-    REAL = 3
+    REAL = auto()
 
     # A scoped name.
-    SCOPED = 4
+    SCOPED = auto()
 
     # A function call.
-    FCALL = 5
+    FCALL = auto()
 
     # A placeholder.
-    EMPTY = 6
-
-
-class ScopedName:
-    """ Encapsulate a scoped name. """
-
-    def __init__(self, name):
-        """ Initialise the scoped name. """
-
-        if isinstance(name, list):
-            # This use is internal to this class.
-            self._name = name
-        elif isinstance(name, ScopedName):
-            # Make a copy of another ScopedName object.
-            self._name = list(name._name)
-        else:
-            # Create a simple name.
-            self._name = [name]
-
-    def __eq__(self, other):
-        """ Compare with another scoped name for equality. """
-
-        if isinstance(other, ScopedName):
-            return self._name == other._name
-
-        return NotImplemented
-
-    def __delitem__(self, index):
-        """ Remove the requested name. """
-
-        del self._name[self._normalised_index(index)]
-
-    def __getitem__(self, index):
-        """ Get the requested name. """
-
-        return self._name[self._normalised_index(index)]
-
-    def __len__(self):
-        """ Return the length of the name, ie. the number of indivual names.
-        """
-
-        nr_names = len(self._name)
-
-        if self.is_absolute:
-            nr_names -= 1
-
-        return nr_names
-
-    def __lt__(self, other):
-        """ Compare with another scoped name to allow lists to be sorted. """
-
-        if isinstance(other, ScopedName):
-            return self._name < other._name
-
-        return NotImplemented
-
-    def __setitem__(self, index, name):
-        """ Set the requested name. """
-
-        self._name[self._normalised_index(index)] = name
-
-    def __str__(self):
-        """ Return the C++ string representation. """
-
-        # The special treatment is a hack to simplify how the result is used.
-        # We ignore any leading '::' and truncate the conversion if we think we
-        # are converting an encoded template name (ie. we find a word that
-        # starts with a digit).
-        s_l = []
-        for s in self:
-            if s[0].isdigit():
-                break
-
-            s_l.append(s)
-
-        return '::'.join(s_l)
-
-    def append(self, name):
-        """ Append a simple name. """
-
-        self._name.append(name)
-
-    @property
-    def base_name(self):
-        """ The base name of the scoped name. """
-
-        return self._name[-1]
-
-    @property
-    def is_absolute(self):
-        """ True if the scoped name is absolute. """
-
-        return self._name[0] == ''
-
-    @property
-    def is_simple(self):
-        """ Return True if the name is simple, ie. relative and unscoped. """
-
-        return len(self) == 1
-
-    def matches(self, scoped_name, scope):
-        """ Return True if a scoped name matches this taking account of a scope
-        if the scoped name is relative.
-        """
-
-        assert self.is_absolute
-
-        # Check for the simple case.
-        if scoped_name.is_absolute:
-            return self == scoped_name
-
-        # Try each scope, inner to outer.
-        while scope is not None:
-            fq_name = ScopedName(scope.iface_file.fq_cpp_name)
-            fq_name._name.extend(scoped_name._name)
-
-            if self == fq_name:
-                return True
-
-            scope = scope.scope
-
-        # Try the global scope.
-        fq_name = ScopedName(scoped_name)
-        fq_name.make_absolute()
-
-        return self == fq_name
-
-    def make_absolute(self):
-        """ Make sure the scoped name is absolute. """
-
-        if self._name[0] != '':
-            self._name.insert(0, '')
-
-    @classmethod
-    def parse(cls, raw):
-        """ Return a ScopedName object by parsing a raw string. """
-
-        return cls(raw.split('::'))
-
-    def prepend(self, scoped_name):
-        """ Prepend a scoped name. """
-
-        new_name = list(scoped_name._name)
-        new_name.extend(self._name)
-        self._name = new_name
-
-    @property
-    def scope(self):
-        """ The scoped name that is the enclosing scope of this one.  It will
-        be None if there isn't one.
-        """
-
-        if len(self) == 1:
-            return None
-
-        return type(self)(self._name[:-1])
-
-    def _normalised_index(self, index):
-        """ Return a normalised index. """
-
-        if not isinstance(index, int):
-            raise TypeError("'{}' is an invalid index".format(type(index)))
-
-        if index >= 0 and self.is_absolute:
-            index += 1
-
-        return index
+    EMPTY = auto()
 
 
 @dataclass
@@ -834,12 +623,20 @@ class Argument:
     # The value of /ScopesStripped/.
     scopes_stripped: int = 0
 
+    # The source location.
+    source_location: Optional['SourceLocation'] = None
+
     # Any transfer of ownership.
     transfer: Transfer = Transfer.NONE
 
     # The non-default type hints.
     type_hints: Optional['TypeHints'] = None
 
+    def __hash__(self):
+        """ Reimplemented so an Argument object can be used as a dict key. """
+
+        return id(self)
+
 
 @dataclass
 class CachedName:
@@ -850,9 +647,20 @@ class CachedName:
     # The name.
     name: str
 
+    # Set if the name is a substring of another. (resolver)
+    is_substring: bool = False
+
+    # The offset of the name in the string pool. (resolver)
+    offset: int = 0
+
     # Set if the name is used in the generated code.
     used: bool = False
 
+    def __str__(self):
+        """ Return the string representation. """
+
+        return self.name
+
 
 @dataclass
 class CodeBlock:
@@ -890,6 +698,9 @@ class Constructor:
     # The action required on the GIL.
     gil_action: GILAction = GILAction.DEFAULT
 
+    # Set if the ctor is the implementation of a cast. (resolver)
+    is_cast: bool = False
+
     # The keyword argument support.
     kw_args: KwArgs = KwArgs.NONE
 
@@ -911,7 +722,7 @@ class Constructor:
     # Set if a Python exception is raised.
     raises_py_exception: bool = False
 
-    # The optional throw arguments.
+    # The optional throw arguments.  Replace with 'noexcept' in SIP v7.
     throw_args: Optional['ThrowArguments'] = None
 
     # Any transfer of ownership.
@@ -939,8 +750,8 @@ class Extract:
     # The order.  A negative value implies the part is appended to the extract.
     order: int
 
-    # The actual part of the extract.
-    part: CodeBlock
+    # The text of the extract part.
+    text: str
 
 
 @dataclass
@@ -962,7 +773,8 @@ class IfaceFile:
     type: IfaceFileType
 
     # The C/C++ name.  It will be None if the interface file relates to a
-    # template.
+    # template.  Note that this is fully qualified and so overlaps with
+    # 'fq_cpp_name'.
     cpp_name: Optional[CachedName] = None
 
     # The filename extension.
@@ -972,6 +784,9 @@ class IfaceFile:
     # relates to a template.
     fq_cpp_name: Optional[ScopedName] = None
 
+    # The generated type number. (resolver)
+    type_nr: int = -1
+
     # The defining module.  It will be None if the interface file relates to a
     # template.
     module: Optional['Module'] = None
@@ -1035,6 +850,15 @@ class MappedType:
     # Set if the handwritten code requires user state information.
     needs_user_state: bool = False
 
+    # Set if /NoAssignmentOperator/ was specified.
+    no_assignment_operator: bool = False
+
+    # Set if /NoCopyCtor/ was specified.
+    no_copy_ctor: bool = False
+
+    # Set if /NoDefaultCtor/ was specified.
+    no_default_ctor: bool = False
+
     # Set if /NoRelease/ was specified.
     no_release: bool = False
 
@@ -1093,6 +917,9 @@ class Member:
     # Set if /Sequence/ was specified.
     is_sequence: bool = False
 
+    # The number of the member. (outputter)
+    member_nr: int = -1
+
     # The original interface file if the function was defined in a namespace.
     namespace_iface_file: Optional[IfaceFile] = None
 
@@ -1107,12 +934,12 @@ class Member:
 class Module:
     """ Encapsulate a module. """
 
+    # The list of all (explicit and implied) imports. (resolver)
+    all_imports: List['Module'] = field(default_factory=list)
+
     # Set if wrapped ctors should support cooperative multi-inheritance.
     call_super_init: bool = False
 
-    # The containing composite module.
-    composite: Optional['Module'] = None
-
     # The text specified by any %Copying directives.
     copying: List[CodeBlock] = field(default_factory=list)
 
@@ -1154,9 +981,6 @@ class Module:
     # The code specified by any %InitialisationCode directives.
     initialisation_code: List[CodeBlock] = field(default_factory=list)
 
-    # Set if the module is a composite module.
-    is_composite: bool = False
-
     # The software license.
     license: Optional[License] = None
 
@@ -1169,9 +993,18 @@ class Module:
     # The next key to auto-allocate.
     next_key: int = -1
 
+    # The number of exceptions defined in this module. (resolver)
+    nr_exceptions: int = 0
+
+    # The generated types needed by this module. (resolver)
+    needed_types: List[Argument] = field(default_factory=list)
+
     # The number of typedefs defined in this module.
     nr_typedefs: int = 0
 
+    # The number of virtual error handlers defined in this module. (resolver)
+    nr_virtual_error_handlers: int = 0
+
     # The overloaded global functions.
     overloads: List['Overload'] = field(default_factory=list)
 
@@ -1184,6 +1017,9 @@ class Module:
     # The code specified by any %PreInitialisationCode directives.
     preinitialisation_code: List[CodeBlock] = field(default_factory=list)
 
+    # The name of the module. (resolver)
+    py_name: Optional[str] = None
+
     # Set if the generated bindings are Py_ssize_t clean.
     py_ssize_t_clean: bool = False
 
@@ -1220,7 +1056,7 @@ class Overload:
     # The member common to all overloads.
     common: Member
 
-    # The C/C++ name.
+    # The C/C++ name if not a operator/slot.
     cpp_name: str
 
     # The C/C++ signature.
@@ -1253,6 +1089,9 @@ class Overload:
     # Set if /AutoGen/ was specified and the associated feature was enabled.
     is_auto_generated: bool = False
 
+    # Set if the overload is a complementary slot. (resolver)
+    is_complementary: bool = False
+
     # Set if the overload is const.
     is_const: bool = False
 
@@ -1262,12 +1101,24 @@ class Overload:
     # Set if the overload is final.
     is_final: bool = False
 
+    # Set if the C++ overload is global. (resolver)
+    is_global: bool = False
+
+    # Set if self should not be dereferenced. (resolver)
+    dont_deref_self: bool = False
+
+    # Set if the overload is a reflected slot. (resolver)
+    is_reflected: bool = False
+
     # Set if the overload is static.
     is_static: bool = False
 
     # Set if the overload is virtual.
     is_virtual: bool = False
 
+    # Set if the overload is a virtual reimplementation. (resolver)
+    is_virtual_reimplementation: bool = False
+
     # The keyword argument support.
     kw_args: KwArgs = KwArgs.NONE
 
@@ -1298,13 +1149,10 @@ class Overload:
     # Set if a Python exception is raised.
     raises_py_exception: bool = False
 
-    # The source file name.
-    source_file: Optional[str] = None
+    # The source location.
+    source_location: Optional['SourceLocation'] = None
 
-    # The source line number.
-    source_line: int = 0
-
-    # The optional throw arguments.
+    # The optional throw arguments.  Replace with 'noexcept' in SIP v7.
     throw_args: Optional['ThrowArguments'] = None
 
     # Any transfer of ownership.
@@ -1368,14 +1216,38 @@ class Signature:
     # The list of arguments.
     args: List[Argument] = field(default_factory=list)
 
-    # The type of the optional result.
+    # The type of the result.
     result: Optional[Argument] = None
 
 
+@dataclass
+class SourceLocation:
+    """ Encapsulate a location in a .sip source file. """
+
+    # The .sip file name.
+    sip_file: str
+
+    # The column number.
+    column: int = 0
+
+    # The line number.
+    line: int = 0
+
+
 @dataclass
 class Specification:
     """ Encapsulate a parsed .sip file. """
 
+    # The version of the ABI being targeted.
+    abi_version: tuple
+
+    # Set if the specification is strict.
+    is_strict: bool
+
+    # The fully qualified name of the sip module.  If it is None then there is
+    # no shared sip module.
+    sip_module: Optional[str]
+
     # Set if the bindings are for C rather than C++.
     c_bindings: bool = False
 
@@ -1400,30 +1272,48 @@ class Specification:
     # The interface files.
     iface_files: List[IfaceFile] = field(default_factory=list)
 
+    # Set if the specification is for a composite module.
+    is_composite: bool = False
+
     # The mapped type templates.
     mapped_type_templates: List[MappedTypeTemplate] = field(default_factory=list)
 
     # The mapped types.
     mapped_types: List[MappedType] = field(default_factory=list)
 
-    # The list of modules.  Initially there is an unnamed module.
-    modules: List[Module] = field(default_factory=lambda: [Module()])
+    # The module for which code is to be generated.
+    module: Module = field(default_factory=Module)
 
     # The cache of names that may be required as strings in the generated code.
-    name_cache: List[CachedName] = field(default_factory=list)
+    name_cache: Dict[int, List[CachedName]] = field(default_factory=dict)
+
+    # The number of virtual handlers. (resolver)
+    nr_virtual_handlers: int = 0
 
     # The list of plugins.  Note that these are PyQt-specific and will be
     # removed in SIP v7.
     plugins: List[str] = field(default_factory=list)
 
+    # The QObject class.
+    pyqt_qobject: Optional['WrappedClass'] = None
+
     # The list of typedefs.
     typedefs: List['WrappedTypedef'] = field(default_factory=list)
 
+    # The list of variables.
+    variables: List['WrappedVariable'] = field(default_factory=list)
+
     # The list of virtual error handlers.
     virtual_error_handlers: List['VirtualErrorHandler'] = field(default_factory=list)
 
-    # The list of variables.
-    variables: List['WrappedVariable'] = field(default_factory=list)
+    # The list of virtual handlers. (resolver)
+    virtual_handlers: List['VirtualHandler'] = field(default_factory=list)
+
+    def __hash__(self):
+        """ Reimplemented so a Specification object can be used as a dict key.
+        """
+
+        return id(self)
 
 
 @dataclass
@@ -1442,7 +1332,7 @@ class ThrowArguments:
     """ Encapsulate the arguments to a C++ throw(). """
 
     # The list of the argument names. If it is None then 'noexcept' was
-    # specified.
+    # specified, otherwise there will be at least one argument.
     arguments: Optional[List['WrappedException']] = None
 
 
@@ -1468,7 +1358,7 @@ class Value:
     value_type: ValueType
 
     # Any literal value.
-    value: Optional[Union[str, int, float, FunctionCall]]
+    value: Optional[Union[str, int, float, FunctionCall, ScopedName]]
 
     # Any binary operator.
     binary_operator: Optional[str] = None
@@ -1493,6 +1383,58 @@ class VirtualErrorHandler:
     # The name of the handler.
     name: str
 
+    # The number of the handler. (resolver)
+    handler_nr: int = -1
+
+
+@dataclass
+class VirtualHandler:
+    """ Encapsulate a virtual overload handler. (resolver) """
+
+    # The C/C++ signature.
+    cpp_signature: Signature
+
+    # The Python signature.
+    py_signature: Signature
+
+    # The code specified by any %VirtualCatcherCode directive.
+    virtual_catcher_code: Optional[CodeBlock]
+
+    # The virtual error handler.
+    virtual_error_handler: VirtualErrorHandler
+
+    # Set if execution should abort if there is an exception.
+    abort_on_exception: bool = False
+
+    # The number of the handler.
+    handler_nr: int = -1
+
+    # Set if ownership of the result should be transferred.
+    transfer_result: bool = False
+
+
+@dataclass
+class VirtualOverload:
+    """ Encapsulate a virtual overloaded member function. (resolver) """
+
+    # The overload
+    overload: Overload
+
+    # The handler for the overload.  It is only set for the module for which
+    # code is being generated.
+    handler: Optional[VirtualHandler]
+
+
+@dataclass
+class VisibleMember:
+    """ Encapsulate a visible member function. (resolver) """
+
+    # The member function.
+    member: Member
+
+    # The defining class.
+    scope: 'WrappedClass'
+
 
 @dataclass
 class WrappedClass:
@@ -1519,6 +1461,9 @@ class WrappedClass:
     # Set if an instance of the class cannot be assigned.
     cannot_assign: bool = False
 
+    # Set if an instance of the class cannot be copied. (resolver)
+    cannot_copy: bool = False
+
     # The list of operator casts.
     casts: List[Argument] = field(default_factory=list)
 
@@ -1558,7 +1503,7 @@ class WrappedClass:
     # The action required on the GIL.
     dtor_gil_action: GILAction = GILAction.DEFAULT
 
-    # The optional dtor throw arguments.
+    # The optional dtor throw arguments.  Replace with 'noexcept' in SIP v7.
     dtor_throw_args: Optional[ThrowArguments] = None
 
     # The code specified by any dtor %VirtualCatcherCode directive.
@@ -1585,6 +1530,12 @@ class WrappedClass:
     # Set if the class has a non-lazy method.
     has_nonlazy_method: bool = False
 
+    # Set if the class actually has a shadow (ie. derived) class. (resolver)
+    has_shadow: bool = False
+
+    # Set if the class has variables that need handlers. (resolver)
+    has_variable_handlers: bool = False
+
     # The %InstanceCode.
     instance_code: Optional[CodeBlock] = None
 
@@ -1600,11 +1551,14 @@ class WrappedClass:
     # Set if the class is opaque.
     is_opaque: bool = False
 
-    # The class is defined in a protected section.
+    # Set if the class is defined in a protected section.
     is_protected: bool = False
 
-    # The class has only been used as a template argument.
-    is_template_arg: bool = False
+    # Set if the class is QObject or a sub-class. (resolver)
+    is_qobject: bool = False
+
+    # The C++ name of any overload annotated with __len__.
+    len_cpp_name: Optional[str] = None
 
     # The methods.
     members: List[Member] = field(default_factory=list)
@@ -1615,6 +1569,16 @@ class WrappedClass:
     # Set if /Mixin/ was specified.
     mixin: bool = False
 
+    # The list of all classes in the class hierarchy starting with itself.
+    # (resolver)
+    mro: List['WrappedClass'] = field(default_factory=list)
+
+    # Set if the class needs an array helper. (resolver)
+    needs_array_helper: bool = False
+
+    # Set if the class needs a copy helper. (resolver)
+    needs_copy_helper: bool = False
+
     # Set if the class needs a shadow (ie. derived) class.
     needs_shadow: bool = False
 
@@ -1652,6 +1616,9 @@ class WrappedClass:
     # The real class if this is a proxy or a namespace extender.
     real_class: Optional['WrappedClass'] = None
 
+    # The sub-class base class. (resolver)
+    subclass_base: Optional['WrappedClass'] = None
+
     # The super-classes.
     superclasses: List['WrappedClass'] = field(default_factory=list)
 
@@ -1673,6 +1640,17 @@ class WrappedClass:
     # The name of the virtual error handler to use.
     virtual_error_handler: Optional[str] = None
 
+    # The virtual overloaded methods. (resolver)
+    virtual_overloads: List[VirtualOverload] = field(default_factory=list)
+
+    # The visible member functions. (resolver)
+    visible_members: List[VisibleMember] = field(default_factory=list)
+
+    def __hash__(self):
+        """ Reimplemented so an Argument object can be used as a dict key. """
+
+        return id(self)
+
 
 @dataclass
 class WrappedEnum:
@@ -1690,7 +1668,7 @@ class WrappedEnum:
     # The cached fully qualified C++ name.
     cached_fq_cpp_name: Optional[CachedName] = None
 
-    # Set if the enum is defined in the protected section.
+    # Set if the enum is defined in a protected section.
     is_protected: bool = False
 
     # Set if the enum is a scoped enum.
@@ -1699,18 +1677,37 @@ class WrappedEnum:
     # The members.
     members: List['WrappedEnumMember'] = field(default_factory=list)
 
+    # Set if this enum is needed by the module for which code is to be
+    # generated. (resolver)
+    needed: bool = False
+
     # Set if /NoScope/ was specified.
     no_scope: bool = False
 
     # Set if the type hint should be suppressed.
     no_type_hint: bool = False
 
+    # The overloaded slot member functions. (resolver)
+    overloads: List['Overload'] = field(default_factory=list)
+
     # The Python name.
     py_name: Optional[CachedName] = None
 
     # The enclosing scope.
     scope: Optional[Union[MappedType, WrappedClass]] = None
 
+    # The slot member functions.  These can only be created by global operators
+    # being moved. (resolver)
+    slots: List[Member] = field(default_factory=list)
+
+    # The generated type number. (resolver)
+    type_nr: int = -1
+
+    def __hash__(self):
+        """ Reimplemented so an Argument object can be used as a dict key. """
+
+        return id(self)
+
 
 @dataclass
 class WrappedEnumMember:
@@ -1740,17 +1737,25 @@ class WrappedException:
     raise_code: CodeBlock
 
     # The base exception if it is a builtin.
-    builtin_base_exception: Optional[str]
+    builtin_base_exception: Optional[str] = None
 
     # The class that implements the exception (if the exception is not a Python
     # exception).
-    class_exception: Optional[WrappedClass]
+    class_exception: Optional[WrappedClass] = None
 
     # The base exception if it is defined in the specification.
-    defined_base_exception: Optional['WrappedException']
+    defined_base_exception: Optional['WrappedException'] = None
+
+    # The number of the exception (only if a Python exception object is
+    # required. (resolver)
+    exception_nr: int = -1
+
+    # Set if this exception is needed by the module for which code is to be
+    # generated. (resolver)
+    needed: bool = False
 
     # The Python name.
-    py_name: Optional[str]
+    py_name: Optional[str] = None
 
 
 @dataclass
@@ -1801,6 +1806,9 @@ class WrappedVariable:
     # Set if the variable is static.
     is_static: bool = False
 
+    # Set if the variable needs a handler. (resolver)
+    needs_handler: bool = False
+
     # Set if the type hint should be suppressed.
     no_type_hint: bool = False
 
